"""**************
Linear Algebra
**************

Linear algebraic functions, calculations of important matrices.
"""

from __future__ import annotations

from .eigenstuff import *
from .fast_bp import *
from .matrices import *

# import helper functions for use in other places
from .resistance import *
