"""Module for graph functions."""

from __future__ import annotations

# Importing the graphfcns modules to register them
from . import (
    design_graphfcns,  # noqa: F401
    network_cleaning_graphfcns,  # noqa: F401
    outfall_graphfcns,  # noqa: F401
    subcatchment_graphfcns,  # noqa: F401
    topology_graphfcns,  # noqa: F401
)
